"""
Provides some utility functions.
"""

from traceback import format_tb
from typing import Callable

from buzz import DoExceptParams
from snick import dedent


def noop(*args, **kwargs):
    """
    This is a no-op function that...does nothing.
    """
    pass


def log_error(logger: Callable[..., None], dep: DoExceptParams):
    """
    Logs an en error with the supplied message, a string representation of the error, and its
    traceback. If the logger supplied is noop, do nothing. Pass as a partial when using the Buzz
    `handle_errors` context manager::

        with Buzz.handle_errors("Boom!", do_except=partial(log_error, debug_logger)):
            do_some_risky_stuff()
    """
    if logger is noop:
        return

    message_template = dedent(
        """
        {final_message}

        Error:
        ______
        {err}

        Traceback:
        ----------
        {trace}
    """
    )

    logger(
        message_template.format(
            final_message=dep.final_message,
            err=str(dep.err),
            trace="\n".join(format_tb(dep.trace)),
        )
    )



from __future__ import annotations

from dataclasses import dataclass

import pendulum
from rich.progress import Progress

from armasec_cli.exceptions import ArmasecCliError


@dataclass
class Tick:
    counter: int
    elapsed: pendulum.Duration
    total_elapsed: pendulum.Duration


class TimeLoop:
    advent: pendulum.DateTime | None
    moment: pendulum.DateTime | None
    last_moment: pendulum.DateTime | None
    counter: int
    progress: Progress | None
    duration: pendulum.Duration
    message: str
    color: str

    def __init__(
        self,
        duration: pendulum.Duration | int,
        message: str = "Processing",
        color: str = "green",
    ):
        self.moment = None
        self.last_moment = None
        self.counter = 0
        self.progress = None
        if isinstance(duration, int):
            ArmasecCliError.require_condition(
                duration > 0,
                "The duration must be a positive integer",
            )
            self.duration = pendulum.duration(seconds=duration)
        else:
            self.duration = duration
        self.message = message
        self.color = color

    def __del__(self):
        """
        Explicitly clear the progress meter if the time-loop is destroyed.
        """
        self.clear()

    def __iter__(self) -> "TimeLoop":
        """
        Start the iterator.

        Creates and starts the progress meter
        """
        self.advent = self.last_moment = self.moment = pendulum.now()
        self.counter = 0
        self.progress = Progress()
        self.progress.add_task(
            f"[{self.color}]{self.message}...",
            total=self.duration.total_seconds(),
        )
        self.progress.start()
        return self

    def __next__(self) -> Tick:
        """
        Iterates the time loop and returns a tick.

        If the duration is complete, clear the progress meter and stop iteration.
        """
        progress: Progress = ArmasecCliError.enforce_defined(
            self.progress,
            "Progress bar has not been initialized...this should not happen",
        )

        self.counter += 1
        self.last_moment = self.moment
        self.moment: pendulum.DateTime = pendulum.now()
        elapsed: pendulum.Duration = self.moment - self.last_moment
        total_elapsed: pendulum.Duration = self.moment - self.advent

        for task_id in progress.task_ids:
            progress.advance(task_id, elapsed.total_seconds())

        if progress.finished:
            self.clear()
            raise StopIteration

        return Tick(
            counter=self.counter,
            elapsed=elapsed,
            total_elapsed=total_elapsed,
        )

    def clear(self):
        """
        Clear the time-loop.

        Stops the progress meter (if set) and reset moments, counter, progress meter.
        """
        if self.progress is not None:
            self.progress.stop()
        self.counter = 0
        self.progress = None
        self.moment = None
        self.last_moment = None



"""
This module provides pydantic schemas for JSON Web Keys.
"""

from typing import List, Optional

from pydantic import ConfigDict, BaseModel


class JWK(BaseModel):
    """
    This Model provides a specification for the objects retrieved from JWK endpoints in OIDC
    providers. It also assists with validation and item access.

    Attributes:
        alg: The algorithm to use for hash validation.
        e:   The exponent parameter to use in RS256 hashing
        kid: The "kid" claim to uniquely identify the key.
        kty: The "kty" claim to identify the type of the key.
        n:   The modulus parameter to use in RS256 hashing.
        use: The claim that identifies the intended use of the public key.
        x5c: The X.509 certificate chain parameter
        x5c: The X.509 certificate SHA-1 thumbprint parameter
    """

    alg: str
    e: str
    kid: str
    kty: str
    n: str

    use: Optional[str] = None
    x5c: Optional[List[str]] = None
    x5t: Optional[str] = None
    model_config = ConfigDict(extra="allow")


class JWKs(BaseModel):
    """
    This Model provides a specification for the container object retrieved from JWK endpoints in
    OIDC providers. It also assists with validation and item access.

    Attributes:
        keys: The list of JWKs contained within.
    """

    keys: List[JWK]



"""
This module defines the core Armasec class.
"""

from functools import lru_cache
from typing import Callable, List, Optional

from fastapi import HTTPException, status

from armasec.schemas import DomainConfig
from armasec.token_security import PermissionMode, TokenSecurity
from armasec.utilities import noop


class Armasec:
    """
    This is a factory class for TokenSecurity. It allows the machinery of armasec to be initialized
    correctly so that the factory method `lockdown` can initialize new instances of TokenSecurity
    to protect routes. It's not essential to use Armasec to secure routes, but it cuts down on the
    boilerplate necessary to do so.
    """

    def __init__(
        self,
        domain_configs: Optional[List[DomainConfig]] = None,
        debug_logger: Optional[Callable[[str], None]] = noop,
        debug_exceptions: bool = False,
        **kargs,
    ):
        """
        Stores initialization values for the TokenSecurity. All are passed through.

        Args:
            domain_configs:   List of domain configuration to authenticate the tokens against.
            debug_logger:     A callable, that if provided, will allow debug logging. Should be
                              passed as a logger method like `logger.debug`
            debug_exceptions: If True, raise original exceptions. Should only be used in a testing
                              or debugging context.
            kargs:            Arguments compatible to instantiate the DomainConfig model.
        """
        primary_domain_config = DomainConfig(**kargs)
        if primary_domain_config.domain:
            self.domain_configs = [primary_domain_config]
        elif domain_configs is not None:
            self.domain_configs = domain_configs
        else:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="No domain was input.",
                headers={"WWW-Authenticate": "Bearer"},
            )
        self.debug_logger = debug_logger
        self.debug_exceptions = debug_exceptions

    @lru_cache(maxsize=128)
    def lockdown(
        self,
        *scopes: str,
        permission_mode: PermissionMode = PermissionMode.ALL,
        skip_plugins: bool = False,
    ) -> TokenSecurity:
        """
        Initialize an instance of TokenSecurity to lockdown a route. Uses memoization to minimize
        the number of TokenSecurity instances initialized. Applies supplied permission_mode when
        checking token permssions against TokenSecurity scopes.

        Args:
            scopes: A list of scopes needed to access the endpoint.
            permissions_mode: If "ALL", all scopes listed are required for access. If "SOME", only
                one of the scopes listed are required for access.
            skip_plugins: If True, do not evaluate plugin validators.
        """
        return TokenSecurity(
            domain_configs=self.domain_configs,
            scopes=scopes,
            permission_mode=permission_mode,
            debug_logger=self.debug_logger,
            debug_exceptions=self.debug_exceptions,
            skip_plugins=skip_plugins,
        )

    def lockdown_all(
        self,
        *scopes: str,
        skip_plugins: bool = False,
    ) -> TokenSecurity:
        """
        Initialize an instance of TokenSecurity to lockdown a route. Uses memoization to minimize
        the number of TokenSecurity instances initialized. Requires all the scopes in the
        TokenSecurity instance to be included in the token permissions. This is just a wrapper
        around `lockdown()` with default permission_mode and is only included for symmetry.

        Args:
            scopes: A list of the scopes needed to access the endpoint. All are required.
            skip_plugins: If True, do not evaluate plugin validators.
        """
        return self.lockdown(
            *scopes,
            permission_mode=PermissionMode.ALL,
            skip_plugins=skip_plugins,
        )

    def lockdown_some(
        self,
        *scopes: str,
        skip_plugins: bool = False,
    ) -> TokenSecurity:
        """
        Initialize an instance of TokenSecurity to lockdown a route. Uses memoization to minimize
        the number of TokenSecurity instances initialized. Requires at least one permission in the
        token to match a scope attached to the TokenSecurity instance.

        Args:
            scopes: A list of the scopes needed to access the endpoint. Only one is required.
            skip_plugins: If True, do not evaluate plugin validators.
        """
        return self.lockdown(
            *scopes,
            permission_mode=PermissionMode.SOME,
            skip_plugins=skip_plugins,
        )



"""
This module defines a TokenManager that can be used to extract token payloads from request headers.
"""

from typing import Callable, Optional, Union

from fastapi.security.utils import get_authorization_scheme_param
from starlette.datastructures import Headers

from armasec.exceptions import AuthenticationError
from armasec.schemas import OpenidConfig
from armasec.token_decoder import TokenDecoder
from armasec.token_payload import TokenPayload
from armasec.utilities import noop


class TokenManager:
    """
    Handle auth via a TokenDecoder and manage extraction from request headers and serialization into
    TokenPayload instances.
    """

    auth_scheme = "bearer"
    header_key = "Authorization"

    def __init__(
        self,
        openid_config: OpenidConfig,
        token_decoder: TokenDecoder,
        audience: Optional[str] = None,
        debug_logger: Optional[Callable[..., None]] = None,
        decode_options_override: Optional[dict] = None,
    ):
        """
        Initialize a base TokenManager.

        Args:
            openid_config:           The openid_configuration needed for claims such as 'issuer'.
            token_decoder:           The decoder used to verify jwts
            audience:                An optional audience to check in decoded tokens.
            debug_logger:            A callable, that if provided, will allow debug logging. Should
                                     be passed as a logger method like `logger.debug`
            decode_options_override: Options that can override the default behavior of the jwt
                                     decode method. For example, one can ignore token expiration by
                                     setting this to `{ "verify_exp": False }`
        """
        self.audience = audience
        self.debug_logger = debug_logger if debug_logger else noop
        self.decode_options_override = decode_options_override if decode_options_override else {}

        self.openid_config = openid_config
        self.token_decoder = token_decoder

    def unpack_token_from_header(self, headers: Union[Headers, dict]) -> str:
        """
        Unpack a JWT from a request header.

        Args:
            headers: The headers from which to unpack a JWT.
        """
        self.debug_logger(f"Attempting to unpack token from headers {headers}")
        auth_str = headers.get(self.header_key)
        self.debug_logger(f"Got {auth_str} using header key {self.header_key}")
        AuthenticationError.require_condition(
            auth_str,
            f"Could not find auth header at {self.header_key}",
        )
        auth_str = str(auth_str)

        self.debug_logger("Attempting to get authorization scheme")
        (scheme, token) = get_authorization_scheme_param(auth_str)
        AuthenticationError.require_condition(
            scheme and token,
            f"Could not extract scheme ('{self.auth_scheme}') from token '{token}'",
        )
        AuthenticationError.require_condition(
            scheme.lower() == self.auth_scheme,
            f"Invalid auth scheme '{scheme}': expected '{self.auth_scheme}'",
        )
        return token

    def extract_token_payload(self, headers: Union[Headers, dict]) -> TokenPayload:
        """
        Retrieve a token from a request header and decode it into a TokenPayload.

        Args:
            headers: The headers from which to retrieve a JWT.
        """
        token = self.unpack_token_from_header(headers)
        token_payload = self.token_decoder.decode(token, audience=self.audience)
        return token_payload



"""
This module defines a pydantic schema for the payload of a jwt.
"""

from datetime import datetime
from typing import List, Optional

from pydantic import ConfigDict, BaseModel, Field, AliasChoices


class TokenPayload(BaseModel):
    """
    A convenience class that can be used to access parts of a decoded jwt.

    Attributes:
        sub:            The "sub" claim from a JWT.
        permissions:    The permissions claims extracted from a JWT.
        expire:         The "exp" claim extracted from a JWT.
        client_id:      The "azp" claim extracted from a JWT.
        original_token: The original token value
    """

    sub: str
    permissions: List[str] = Field(list())
    expire: datetime = Field(None, validation_alias=AliasChoices("exp", "expire"))
    client_id: str = Field(None, validation_alias=AliasChoices("azp", "client_id"))
    original_token: Optional[str] = None
    model_config = ConfigDict(extra="allow")

    def to_dict(self):
        """
        Convert a TokenPayload to the equivalent dictionary returned by `jwt.decode()`.
        """
        return dict(
            sub=self.sub,
            permissions=self.permissions,
            exp=int(self.expire.timestamp()),
            client_id=self.client_id,
        )



"""
This module provides a pydantic schema describing Armasec's configuration parameters.
"""

from typing import Any, Dict, List, Optional, Set, Union

import snick
from pydantic import BaseModel, Field


class DomainConfig(BaseModel):
    """
    This model provides a specification for the input domains to authenticate against.
    It expects the domain indeed and the audience to refer to.

    Attributes:
        domain:     The OIDC domain from which resources are loaded.
        audience:   Optional designation of the token audience.
        algorithm:  The Algorithm to use for decoding. Defaults to RS256.
        use_https:  If true, use `https` for URLs. Otherwise use `http`
        match_keys: Dictionary of k/v pairs to match in the token when decoding it.
    """

    domain: str = Field(str(), description="The OIDC domain where resources are loaded.")
    audience: Optional[str] = Field(None, description="Optional designation of the token audience.")
    algorithm: str = Field(
        "RS256", description="The the algorithm to use for decoding. Defaults to RS256."
    )
    use_https: bool = Field(
        True,
        description=snick.unwrap(
            """
            If falsey, use ``http`` when pulling openid config from the OIDC server
            instead of ``https`` (the default).
            """,
        ),
    )
    match_keys: Dict[str, Union[str, List[Any], Dict[Any, Any], Set[Any], bool, int, float]] = (
        Field(
            dict(),
            description=snick.unwrap(
                """
            Dictionary of key-value pair to match in the token when decoding it. It will
            raise 403 in case the input key-value pair cannot be found in the token.
            """
            ),
        )
    )
    payload_claim_mapping: Optional[Dict[str, Any]] = Field(
        None,
        description=snick.unwrap(
            """
            Optional mappings that are applied to map claims to top-level properties of
            TokenPayload. See docs for `TokenDecoder` for more info.
            """
        ),
    )



"""
Hook specification module for armasec plugins.
"""

from typing import Callable

import pluggy
from starlette.requests import Request

from armasec.token_payload import TokenPayload

hookspec = pluggy.HookspecMarker("armasec")


@hookspec
def armasec_plugin_check(
    request: Request,
    token_payload: TokenPayload,
    debug_logger: Callable[..., None],
) -> None:
    """
    Check a token payload for validity against a request using your plugin.

    If the check fails, it should raise a ArmasecError or a subclass thereof.

    Args:
        request:       The original request made to the secured endpoint. Will be passed
                       to the plugin method if the implementation includes it as a
                       keyword argument.
        token_payload: The contents of the auth token. Will be passed ot the plugin
                       method if the implementation includes it as a keyword argument
        debug_logger:  A callable, that if provided, will allow debug logging. Should be
                       passed as a logger method like `logger.debug`. Will be passed to
                       the plugin method if the implementation includes it as a keyword
                       argument.
    """



from __future__ import annotations

from pathlib import Path
from functools import wraps

from loguru import logger

from armasec_cli.schemas import TokenSet
from armasec_cli.exceptions import Abort
from armasec_cli.config import cache_dir


def init_cache(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            cache_dir.mkdir(exist_ok=True, parents=True)
            token_dir = cache_dir / "token"
            token_dir.mkdir(exist_ok=True)
            info_file = cache_dir / "info.txt"
            info_file.write_text("This directory is used by Armasec CLI for its cache.")
        except Exception:
            raise Abort(
                """
                Cache directory {cache_dir} doesn't exist, is not writable, or could not be created.

                Please check your home directory permissions and try again.
                """,
                subject="Non-writable cache dir",
                log_message="Non-writable cache dir",
            )
        return func(*args, **kwargs)

    return wrapper


def _get_token_paths() -> tuple[Path, Path]:
    token_dir = cache_dir / "token"
    access_token_path: Path = token_dir / "access.token"
    refresh_token_path: Path = token_dir / "refresh.token"
    return (access_token_path, refresh_token_path)


def load_tokens_from_cache() -> TokenSet:
    """
    Loads an access token (and a refresh token if one exists) from the cache.
    """
    (access_token_path, refresh_token_path) = _get_token_paths()

    Abort.require_condition(
        access_token_path.exists(),
        "Please login with your auth token first using the `armasec login` command",
        raise_kwargs=dict(subject="You need to login"),
    )

    logger.debug("Retrieving access token from cache")
    token_set: TokenSet = TokenSet(access_token=access_token_path.read_text())

    if refresh_token_path.exists():
        logger.debug("Retrieving refresh token from cache")
        token_set.refresh_token = refresh_token_path.read_text()

    return token_set


def save_tokens_to_cache(token_set: TokenSet):
    """
    Saves tokens from a token_set to the cache.
    """
    (access_token_path, refresh_token_path) = _get_token_paths()

    logger.debug(f"Caching access token at {access_token_path}")
    access_token_path.write_text(token_set.access_token)
    access_token_path.chmod(0o600)

    if token_set.refresh_token is not None:
        logger.debug(f"Caching refresh token at {refresh_token_path}")
        refresh_token_path.write_text(token_set.refresh_token)
        refresh_token_path.chmod(0o600)


def clear_token_cache():
    """
    Clears the token cache.
    """
    logger.debug("Clearing cached tokens")
    (access_token_path, refresh_token_path) = _get_token_paths()

    logger.debug(f"Removing access token at {access_token_path}")
    if access_token_path.exists():
        access_token_path.unlink()

    logger.debug(f"Removing refresh token at {refresh_token_path}")
    if refresh_token_path.exists():
        refresh_token_path.unlink()



from __future__ import annotations


import jose
import pyperclip
import snick
import typer

from armasec_cli.exceptions import Abort, handle_abort
from armasec_cli.schemas import TokenSet, Persona, CliContext
from armasec_cli.cache import init_cache, load_tokens_from_cache, clear_token_cache
from armasec_cli.format import terminal_message, render_json
from armasec_cli.auth import fetch_auth_tokens, extract_persona
from armasec_cli.config import (
    OidcProvider,
    attach_settings,
    init_settings,
    dump_settings,
    clear_settings,
)
from armasec_cli.client import attach_client
from armasec_cli.logging import init_logs


app = typer.Typer()


@app.callback(invoke_without_command=True)
@handle_abort
def main(
    ctx: typer.Context,
    verbose: bool = typer.Option(False, help="Enable verbose logging to the terminal"),
):
    """
    Welcome to the Armasec CLI!

    More information can be shown for each command listed below by running it with the
    --help option.
    """

    if ctx.invoked_subcommand is None:
        terminal_message(
            snick.conjoin(
                "No command provided. Please check [bold magenta]usage[/bold magenta]",
                "",
                f"[yellow]{ctx.get_help()}[/yellow]",
            ),
            subject="Need an Armasec command",
        )
        raise typer.Exit()

    init_logs(verbose=verbose)
    ctx.obj = CliContext()


@app.command()
@handle_abort
@init_cache
@attach_settings
@attach_client
def login(ctx: typer.Context):
    token_set: TokenSet = fetch_auth_tokens(ctx.obj)
    persona: Persona = extract_persona(token_set)
    terminal_message(
        f"User was logged in with email '{persona.identity_data.email}'",
        subject="Logged in!",
    )


@app.command()
@handle_abort
@init_cache
def logout():
    """
    Logs out of the jobbergate-cli. Clears the saved user credentials.
    """
    clear_token_cache()
    terminal_message(
        "User was logged out.",
        subject="Logged out",
    )


@app.command()
@init_cache
def set_config(
    domain: str = typer.Option(..., help="The domain used by your OIDC provider"),
    audience: str = typer.Option(
        ...,
        help="The audience required by your OIDC provider",
    ),
    client_id: str = typer.Option(
        ...,
        help="The unique client ID to use for logging in via CLI",
    ),
    use_https: bool = typer.Option(True, help="Use https for requests"),
    max_poll_time: int = typer.Option(
        5 * 60,
        help="The max time to wait for login to complete",
    ),
    provider: OidcProvider = typer.Option(
        OidcProvider.KEYCLOAK,
        help="The OIDC provider you use",
    ),
):
    settings = init_settings(
        oidc_domain=domain,
        oidc_audience=audience,
        oidc_client_id=client_id,
        oidc_use_https=use_https,
        oidc_max_poll_time=max_poll_time,
        oidc_provider=provider,
    )
    dump_settings(settings)


@app.command()
@handle_abort
@init_cache
@attach_settings
def show_config(ctx: typer.Context):
    """
    Show the current config.
    """
    render_json(ctx.obj.settings.model_dump())


@app.command()
@handle_abort
@init_cache
def clear_config():
    """
    Show the current config.
    """
    clear_settings()


@app.command()
@handle_abort
@init_cache
def show_token(
    plain: bool = typer.Option(
        False,
        help="Show the token in plain text.",
    ),
    refresh: bool = typer.Option(
        False,
        help="Show the refresh token instead of the access token.",
    ),
    show_prefix: bool = typer.Option(
        False,
        "--prefix",
        help="Include the 'Bearer' prefix in the output.",
    ),
    show_header: bool = typer.Option(
        False,
        "--header",
        help="Show the token as it would appear in a request header.",
    ),
    decode: bool = typer.Option(
        False,
        "--decode",
        help="Show the content of the decoded access token.",
    ),
):
    """
    Show the token for the logged in user.

    Token output is automatically copied to your clipboard.
    """
    token_set: TokenSet = load_tokens_from_cache()
    token: str | None = None
    if not refresh:
        token = token_set.access_token
        subject = "Access Token"
        Abort.require_condition(
            token is not None,
            "User is not logged in. Please log in first.",
            raise_kwargs=dict(
                subject="Not logged in",
            ),
        )
    else:
        token = token_set.refresh_token
        subject = "Refresh Token"
        Abort.require_condition(
            token is not None,
            snick.unwrap(
                """
                User is not logged in or does not have a refresh token.
                Please try loggin in again.
                """
            ),
            raise_kwargs=dict(
                subject="No refresh token",
            ),
        )

    if decode:
        # Decode the token with ALL verification turned off (we just want to unpack it)
        content = jose.jwt.decode(
            token,
            "secret-will-be-ignored",
            options=dict(
                verify_signature=False,
                verify_aud=False,
                verify_iat=False,
                verify_exp=False,
                verify_nbf=False,
                verify_iss=False,
                verify_sub=False,
                verify_jti=False,
                verify_at_hash=False,
            ),
        )
        render_json(content)
        return

    if show_header:
        token_text = f"""{{ "Authorization": "Bearer {token}" }}"""
    else:
        prefix = "Bearer " if show_prefix else ""
        token_text = f"{prefix}{token}"

    try:
        pyperclip.copy(token_text)
        on_clipboard = True
    except Exception:
        on_clipboard = False

    if plain:
        print(token_text)
    else:
        kwargs = dict(subject=subject, indent=False)
        if on_clipboard:
            kwargs["footer"] = "The output was copied to your clipboard"

        terminal_message(token_text, **kwargs)



"""
This module defines a TokenSecurity injectable that can be used enforce access on FastAPI routes.
"""

from typing import Callable, Iterable, List, Optional

from auto_name_enum import AutoNameEnum, auto
from fastapi import HTTPException, status
from fastapi.openapi.models import APIKey, APIKeyIn
from fastapi.security.api_key import APIKeyBase
from pydantic import ConfigDict, BaseModel
from snick import unwrap
from starlette.requests import Request

from armasec.exceptions import AuthenticationError, AuthorizationError
from armasec.openid_config_loader import OpenidConfigLoader
from armasec.pluggable import plugin_manager
from armasec.schemas import DomainConfig
from armasec.token_decoder import TokenDecoder
from armasec.token_manager import TokenManager
from armasec.token_payload import TokenPayload
from armasec.utilities import noop


class ManagerConfig(BaseModel):
    """
    Model class to represent a TokenManager instance and its domain configuration for easier mapping

    Attributes:
        manager: The TokenManager instance to use for decoding tokens.
        domain_config: The DomainConfig for the openid server.
    """

    manager: TokenManager
    domain_config: DomainConfig
    model_config = ConfigDict(arbitrary_types_allowed=True)


class PermissionMode(AutoNameEnum):
    """
    Endpoint permissions.

    Attributes:
        ALL:  Require all listed permissions.
        SOME: Require at least one of the listed permissions.
    """

    ALL = auto()
    SOME = auto()


class TokenSecurity(APIKeyBase):
    """
    An injectable Security class that returns a TokenPayload when used with Depends().

    Attributes:
        manager: The TokenManager to use for token validation and extraction.
    """

    manager: Optional[TokenManager]

    def __init__(
        self,
        domain_configs: List[DomainConfig],
        scopes: Optional[Iterable[str]] = None,
        permission_mode: PermissionMode = PermissionMode.ALL,
        debug_logger: Optional[Callable[..., None]] = None,
        debug_exceptions: bool = False,
        skip_plugins: bool = False,
    ):
        """
        Initializes the TokenSecurity instance.

        Args:
            domain_configs:   List of domain configuration to authenticate the tokens against.
            scopes:           Optional permissions scopes that should be checked
            permission_mode:  The PermissionMode to apply in the protected route.
            debug_logger:     A callable, that if provided, will allow debug logging. Should be
                              passed as a logger method like `logger.debug`
            debug_exceptions: If True, raise original exceptions. Should only be used in a testing
                              or debugging context.
            skip_plugins:     If True, do not evaluate plugin validators.
        """
        self.domain_configs = domain_configs
        self.scopes = scopes
        self.permission_mode = permission_mode

        self.debug_logger = debug_logger if debug_logger else noop
        self.debug_exceptions = debug_exceptions
        self.skip_plugins = skip_plugins

        # Settings needed for FastAPI's APIKeyBase
        self.model: APIKey = APIKey(
            **{"in": APIKeyIn.header},  # type: ignore[arg-type]
            name=TokenManager.header_key,
            description=self.__class__.__doc__,
        )
        self.scheme_name = self.__class__.__name__

        # This will be lazy loaded at the first request call
        self.managers: List[ManagerConfig] = list()

    async def __call__(self, request: Request) -> TokenPayload:
        """
        This method is called by FastAPI's dependency injection system when a TokenSecurity instance
        is injected to a route endpoint via the Depends() method. Lazily loads the OIDC config,
        the TokenDecoder, and the TokenManager if they are not already initialized.

        Args:
            request: The FastAPI request to check for secure access.
        """

        try:
            token_payload = self._extract_token_payload_from_manager(request)
        except AttributeError:
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail="Not authorized",
                headers={"WWW-Authenticate": "Bearer"},
            )
        except Exception as err:
            if self.debug_exceptions:
                raise err
            else:
                raise HTTPException(
                    status_code=getattr(err, "status_code", status.HTTP_401_UNAUTHORIZED),
                    detail=getattr(err, "detail", "Not authenticated"),
                    headers={"WWW-Authenticate": "Bearer"},
                )

        if self.scopes:
            token_permissions = set(token_payload.permissions)
            my_permissions = set(self.scopes)

            self.debug_logger(
                unwrap(
                    f"""
                    Checking my permissions {my_permissions} against token_permissions
                    {token_permissions} using PermissionMode {self.permission_mode}
                    """
                )
            )
            try:
                if self.permission_mode == PermissionMode.ALL:
                    message = unwrap(
                        f"""
                        Token permissions {token_permissions} missing some required permissions
                        {my_permissions - token_permissions}
                        """
                    )
                    AuthorizationError.require_condition(
                        my_permissions - token_permissions == set(),
                        message,
                    )
                elif self.permission_mode == PermissionMode.SOME:
                    message = unwrap(
                        f"""
                        Token permissions {token_permissions} missing at least
                        one required permissions
                        {my_permissions}
                        """
                    )
                    AuthorizationError.require_condition(
                        token_permissions & my_permissions,
                        message,
                    )
                else:
                    raise AuthorizationError(f"Unknown permission_mode: {self.permission_mode}")

            except Exception as err:
                if self.debug_exceptions:
                    raise err
                else:
                    raise HTTPException(
                        status_code=getattr(err, "status_code", status.HTTP_403_FORBIDDEN),
                        detail=getattr(err, "detail", "Not authorized"),
                        headers={"WWW-Authenticate": "Bearer"},
                    )

        if not self.skip_plugins:
            self.debug_logger("Applying plugin checks")
            try:
                plugin_manager.hook.armasec_plugin_check(
                    request=request,
                    token_payload=token_payload,
                    debug_logger=self.debug_logger,
                )
            except Exception as err:
                if self.debug_exceptions:
                    raise err
                else:
                    raise HTTPException(
                        status_code=getattr(err, "status_code", status.HTTP_403_FORBIDDEN),
                        detail=getattr(err, "detail", "Not authorized"),
                        headers={"WWW-Authenticate": "Bearer"},
                    )

        return token_payload

    def _load_all_managers(self) -> None:
        if len(self.managers) == 0:
            for domain_config in self.domain_configs:
                try:
                    manager = self._load_manager(domain_config)
                    self.managers.append(
                        ManagerConfig(manager=manager, domain_config=domain_config)
                    )
                except AuthenticationError:
                    self.debug_logger(f"Failed to match JWK against domain {domain_config.domain}")
                except Exception as err:
                    if self.debug_exceptions:
                        self.debug_logger(f"Exception caught: {err.__class__.__name__}")

        AuthenticationError.require_condition(
            len(self.managers) > 0,
            "Not authenticated: couldn't load any TokenManager instance",
        )

    def _load_manager(self, domain_config: DomainConfig) -> TokenManager:
        self.debug_logger(f"Lazy loading TokenManager for domain {domain_config.domain}")
        loader = OpenidConfigLoader(
            domain_config.domain, use_https=domain_config.use_https, debug_logger=self.debug_logger
        )
        decoder = TokenDecoder(
            loader.jwks,
            domain_config.algorithm,
            debug_logger=self.debug_logger,
            payload_claim_mapping=domain_config.payload_claim_mapping,
        )
        return TokenManager(
            loader.config,
            decoder,
            audience=domain_config.audience,
            debug_logger=self.debug_logger,
        )

    def _extract_token_payload_from_manager(self, request: Request) -> TokenPayload:
        token_payload = None

        self._load_all_managers()

        for manager_config in self.managers:
            try:
                token_payload = manager_config.manager.extract_token_payload(request.headers)
            except Exception as err:
                self.debug_logger(f"Exception caught: {err.__class__.__name__}")
            else:
                AuthenticationError.require_condition(
                    token_payload is not None,
                    "Not authenticated: could not find matching JWK"
                    " with any input domain or token is malformed",
                )
                message = "Not authorized: token doesn't contain necessary key-value pairs"
                for key_to_match, value_to_match in manager_config.domain_config.match_keys.items():
                    if isinstance(value_to_match, bool):
                        AuthorizationError.require_condition(
                            getattr(token_payload, key_to_match) is value_to_match, message
                        )
                    elif isinstance(value_to_match, (str, int, float)):
                        AuthorizationError.require_condition(
                            getattr(token_payload, key_to_match) == value_to_match,
                            message,
                        )
                    else:
                        AuthorizationError.require_condition(
                            set(getattr(token_payload, key_to_match)) & set(value_to_match),
                            message,
                        )
                break

        # make static type analyzer happy :)
        assert token_payload is not None

        return token_payload



import json
from typing import Any

import snick
from rich.console import Console
from rich.panel import Panel


def terminal_message(message, subject=None, color="green", footer=None, indent=True):
    panel_kwargs = dict(padding=1)
    if subject is not None:
        panel_kwargs["title"] = f"[{color}]{subject}"
    if footer is not None:
        panel_kwargs["subtitle"] = f"[dim italic]{footer}[/dim italic]"
    text = snick.dedent(message)
    if indent:
        text = snick.indent(text, prefix="  ")
    console = Console()
    console.print()
    console.print(Panel(text, **panel_kwargs))
    console.print()


def render_json(data: Any):
    console = Console()
    console.print()
    console.print_json(json.dumps(data))
    console.print()



from armasec.armasec import Armasec
from armasec.openid_config_loader import OpenidConfigLoader
from armasec.token_decoder import TokenDecoder
from armasec.token_manager import TokenManager
from armasec.token_payload import TokenPayload
from armasec.token_security import TokenSecurity

__all__ = [
    "Armasec",
    "TokenManager",
    "TokenSecurity",
    "TokenPayload",
    "TokenDecoder",
    "OpenidConfigLoader",
]



"""
This module provides a pydantic schema describing openid-configuration data.
"""

from pydantic import ConfigDict, AnyHttpUrl, BaseModel


class OpenidConfig(BaseModel):
    """
    Provides a specification for the objects retrieved from openid_configuration endpoint of the
    OIDC providers. Only includes needed fields for supported Manager instances. Assists with
    validation and item access.

    Attributes:
        issuer:   The URL of the issuer of the tokens.
        jwks_uri: The URI where JWKs can be foun don the OpenID server.
    """

    issuer: AnyHttpUrl
    jwks_uri: AnyHttpUrl
    model_config = ConfigDict(extra="allow")



from armasec.schemas.armasec_config import DomainConfig
from armasec.schemas.jwks import JWK, JWKs
from armasec.schemas.openid_config import OpenidConfig

__all__ = [
    "DomainConfig",
    "JWK",
    "JWKs",
    "OpenidConfig",
]



import json
from functools import wraps
from pathlib import Path

import snick
import typer
from auto_name_enum import AutoNameEnum, auto
from loguru import logger
from pydantic import BaseModel, ValidationError

from armasec_cli.exceptions import Abort

cache_dir: Path = Path.home() / ".local/share/armasec-cli"
settings_path = cache_dir / "armasec.json"


class OidcProvider(AutoNameEnum):
    AUTH0 = auto()
    KEYCLOAK = auto()


class Settings(BaseModel):
    oidc_domain: str
    oidc_audience: str
    oidc_client_id: str
    oidc_use_https: bool = True
    oidc_max_poll_time: int = 5 * 60  # 5 minutes
    oidc_provider: OidcProvider = OidcProvider.KEYCLOAK


def init_settings(**settings_values):
    try:
        logger.debug("Validating settings")
        return Settings(**settings_values)
    except ValidationError as err:
        raise Abort(
            snick.conjoin(
                "A configuration error was detected.",
                "",
                "Details:",
                "",
                f"[red]{err}[/red]",
            ),
            subject="Configuration Error",
            log_message="Configuration error",
        )


def attach_settings(func):
    @wraps(func)
    def wrapper(ctx: typer.Context, *args, **kwargs):
        try:
            logger.debug(f"Loading settings from {settings_path}")
            settings_values = json.loads(settings_path.read_text())
        except FileNotFoundError:
            raise Abort(
                f"""
                No settings file found at {settings_path}!

                Run the set-config sub-command first to establish your OIDC settings.
                """,
                subject="Settings file missing!",
                log_message="Settings file missing!",
            )
        logger.debug("Binding settings to CLI context")
        ctx.obj.settings = init_settings(**settings_values)
        return func(ctx, *args, **kwargs)

    return wrapper


def dump_settings(settings: Settings):
    logger.debug(f"Saving settings to {settings_path}")
    settings_values = json.dumps(settings.model_dump())
    settings_path.write_text(settings_values)


def clear_settings():
    logger.debug(f"Removing saved settings at {settings_path}")
    settings_path.unlink(missing_ok=True)



import buzz
import starlette.status


class ArmasecError(buzz.Buzz):
    """
    A custom exception class used for checking conditions and handling other exceptions.

    Attributes:
        status_code: The HTTP status code indicated by the error. Set to 400.
    """

    status_code: int = starlette.status.HTTP_400_BAD_REQUEST
    detail: str = "Bad request"


class AuthenticationError(ArmasecError):
    """
    Indicates a failure to authenticate and decode jwt.

    Attributes:
        status_code: The HTTP status code indicated by the error. Set to 401.
    """

    status_code: int = starlette.status.HTTP_401_UNAUTHORIZED
    detail: str = "Not authenticated"


class AuthorizationError(ArmasecError):
    """
    Indicates that the provided claims don't match the claims required for a protected endpoint.

    Attributes:
        status_code: The HTTP status code indicated by the error. Set to 403.
    """

    status_code: int = starlette.status.HTTP_403_FORBIDDEN
    detail: str = "Not authorized"


class PayloadMappingError(ArmasecError):
    """
    Indicates that the configured payload_claim_mapping did not match a path in the token.

    Attributes:
        status_code: The HTTP status code indicated by the error. Set to 500.
    """

    status_code: int = starlette.status.HTTP_500_INTERNAL_SERVER_ERROR
    detail: str = "Server error"



"""
This module provides a pytest plugin for testing.
"""

from collections import namedtuple
from contextlib import _GeneratorContextManager, contextmanager
from datetime import datetime
from typing import Callable, Optional

import httpx
import pytest
import respx
import starlette
from jose import jwt
from snick import dedent, strip_whitespace

from armasec.openid_config_loader import OpenidConfigLoader
from armasec.schemas.armasec_config import DomainConfig
from armasec.schemas.jwks import JWK, JWKs
from armasec.schemas.openid_config import OpenidConfig


@pytest.fixture()
def rs256_domain():
    """
    Provide a fixture that returns a domain for use in other fixtures.

    The value here doesn't really have anything to do with an actual domain name.
    """
    return "armasec.dev"


@pytest.fixture()
def rs256_domain_config(rs256_domain):
    """
    Provide a fixture that returns the DomainConfig model for the default rs256 domain.

    Args:
        rs256_domain: An implicit fixture parameter.
    """
    return DomainConfig(domain=rs256_domain, audience="https://this.api")


@pytest.fixture()
def rs256_iss(rs256_domain):
    """
    Provide a fixture that returns an issuer claim for use in other fixtures.

    Args:
        rs256_domain: An implicit fixture parameter.
    """
    return f"https://{rs256_domain}"


@pytest.fixture()
def rs256_kid():
    """
    Provide a fixture that returns a KID header value for use in other fixtures.
    """
    return "SAMPLE_KID"


@pytest.fixture()
def rs256_sub():
    """
    Provide a fixture that returns a sum claim for use in other fixtures.
    """
    return "SAMPLE_SUB"


@pytest.fixture()
def rs256_private_key():
    """
    Provide a fixture that returns a pre-generated private key for RS256 hashing in other fixtures.
    """
    return dedent(
        """
        -----BEGIN RSA PRIVATE KEY-----
        MIIEpAIBAAKCAQEAw408+QDZ10idz4ytJtwFQE4YgmrjvCoEXjtTUWQ3H4nWAAYQ
        +oE9xpr/gosNiFMuyRburvXT+Rkq8ry8tWoUzN2zViaarot+Tt9I71sVlnIsbtDZ
        +XrteMvBwjARn/MEAQEwDLvVzrBnAZrOTwrIkznyJttZh7STrt6y5X91i2MMm3xu
        9QK90kpu3rymAyT5V+AEIRzZai/ZT4YfLDutXulOVlWPQ55Xww1mbheGQ99fUMo5
        LmkxM5Jsz8ulIVvq/G/8guiKwAPJN/8S34NbkgL5GoeXT8uNDkbhtkLh5+o2T4EL
        9/ODKHqx46pHgUmBiC6wNv6uJXdH7qpaqhPR3QIDAQABAoIBABeyl/788Wk7bZRn
        UdxxsVk3nZTAa1S0Ks9YlSI56MwzofFiys/wtZHJ2sjxHPS2T+cilk4xkDyRpjjA
        UoYRku+4tjDsgLZCRU49lNMc0KLotyW+vYuUMA8BcjucI6akhomwoSgJ40Em83So
        U/QUNHZTAVtgHZtqcLMyXa+eIJqBcfsMHFkCgSF8LSD/XkRBMm1SREswDw6KqQQ0
        sZ/8TVF9sJTi3/OG8m5OfI+44AYDaMH5wKoOBcR3FBln+dEutB6JuRjmpnEjQpIT
        DggULc+Dzb/c75yhT1qZSEL3Z99JQTbytPm6boNGKmzUE9HCoY84wKfnhUDocFKW
        jnHMmKkCgYEA7/gRAtLjbJW1rbxw8xN3cyOZEJsMFt4mXMmne6nDTttVKb0wUuXJ
        H8prKAXDOzadAZgPeGJXVSNgGoNeNtkmEKDtysrRiZbWiTRxYPE36MrHFGOywjTn
        tP8qMJHmHYkxS16nqrOl0znUWv6Q6/qwd59Utuu4IJF/CxqP3Z6HPssCgYEA0J2M
        1gRgGj8NnGoIKS58gc3Aa5RdqWKoeiyXeN/zRDfMCKpPsVykvZJb4cLcEdcwe9kC
        3xpgIPaTZCPwhJ1rYiZ0/Xr7oIf0E66IeEKs/bchKcT9+sSaWgc5/zQ7aQ/XpwzU
        nKCTTeMGFUyulCIkoe2tLEQ+Mw1OphIXv17fNPcCgYEAjWgVxh81ivgRjiZ8PJEd
        E4lHmmRzVEpmOslN225nO+G9ppHolwD3ardiO7xhllQRYy4S97KjmfT1ncoJy7Jc
        XvImDhlELprnIwT3RtP+STys4ZP6c7yvSZYPa32eJ4t/s9U8YjfooLb0LwbRqW0Z
        bfRC/GOdJfv27Dkjy8muEs8CgYEAo+oHDOonMLg2U54kh2cVQVCPTngnF76DLmv3
        IGym0gUddfmL4Iowjxt+wma/T+1LFSSwUuiAe6YCrX5nr2uZQmeBKOIG8F2idAyB
        Ai0xi7Dmh9FW1kDAHtjqwxEhVS2zfnhgXij1VQ96aiX0TkR9kBYWKV/9l1NvZqF0
        s1MyAoUCgYA0wfJrCTXdWitkyfxApcmoTxt0ljqUwO6F5fhojf8PU1ouglgkRXtm
        1rSDGp7YUfODhWNSsN2P/eaDybcZo+TGtLQJ5Bai3Qxqh8xPaKCsSZbcPRLRP0w5
        CbTvFEyj6EBEH+TJL/Loa4hKFuAk7ErBAtzMCw6LchTjB/OF+dUusA==
        -----END RSA PRIVATE KEY-----
        """
    ).encode("utf-8")


@pytest.fixture()
def rs256_public_key():
    """
    Provide a fixture that returns a pre-generated public key for RS256 hashing in other fixtures.
    """
    return dedent(
        """
        -----BEGIN RSA PUBLIC KEY-----
        MIIBCgKCAQEAw408+QDZ10idz4ytJtwFQE4YgmrjvCoEXjtTUWQ3H4nWAAYQ+oE9
        xpr/gosNiFMuyRburvXT+Rkq8ry8tWoUzN2zViaarot+Tt9I71sVlnIsbtDZ+Xrt
        eMvBwjARn/MEAQEwDLvVzrBnAZrOTwrIkznyJttZh7STrt6y5X91i2MMm3xu9QK9
        0kpu3rymAyT5V+AEIRzZai/ZT4YfLDutXulOVlWPQ55Xww1mbheGQ99fUMo5Lmkx
        M5Jsz8ulIVvq/G/8guiKwAPJN/8S34NbkgL5GoeXT8uNDkbhtkLh5+o2T4EL9/OD
        KHqx46pHgUmBiC6wNv6uJXdH7qpaqhPR3QIDAQAB
        -----END RSA PUBLIC KEY-----
        """
    ).encode("utf-8")


@pytest.fixture()
def rs256_jwk(rs256_kid):
    """
    Provide a fixture that returns a JWK for use in other fixtures.

    Args:
        rs256_kid: An implicit fixture parameter.
    """
    return JWK(
        alg="RS256",
        kty="RSA",
        kid=rs256_kid,
        n=strip_whitespace(
            """
                w408-QDZ10idz4ytJtwFQE4YgmrjvCoEXjtTUWQ3H4nWAAYQ-oE9xpr_gosNiFMuyRburvXT-Rkq8ry8tWoU
                zN2zViaarot-Tt9I71sVlnIsbtDZ-XrteMvBwjARn_MEAQEwDLvVzrBnAZrOTwrIkznyJttZh7STrt6y5X91
                i2MMm3xu9QK90kpu3rymAyT5V-AEIRzZai_ZT4YfLDutXulOVlWPQ55Xww1mbheGQ99fUMo5LmkxM5Jsz8ul
                IVvq_G_8guiKwAPJN_8S34NbkgL5GoeXT8uNDkbhtkLh5-o2T4EL9_ODKHqx46pHgUmBiC6wNv6uJXdH7qpa
                qhPR3Q
            """
        ),
        e="AQAB",
    )


@pytest.fixture()
def build_rs256_token(rs256_private_key, rs256_iss, rs256_sub, rs256_kid):
    """
    Provide a fixture that returns a helper method that can build a JWT.

    The JWT is signed with the private key provided by the rs256_private_key.

    Args:
        rs256_private_key: An implicit fixture parameter.
        rs256_iss:         An implicit fixture parameter.
        rs256_sub:         An implicit fixture parameter.
    """
    base_claims = dict(
        iss=rs256_iss,
        sub=rs256_sub,
        permissions=[],
    )
    base_headers = dict(kid=rs256_kid)

    def _helper(
        claim_overrides: Optional[dict] = None,
        headers_overrides: Optional[dict] = None,
    ):
        """
        Encode a jwt token with the default claims and headers overridden with user supplied values.
        """
        if claim_overrides is None:
            claim_overrides = dict()

        if headers_overrides is None:
            headers_overrides = dict()

        now = int(datetime.utcnow().timestamp())

        return jwt.encode(
            {
                "iat": now,
                "exp": now + 60 * 60,  # expires in an hour from now
                **base_claims,
                **claim_overrides,
            },
            rs256_private_key,
            algorithm="RS256",
            headers={
                **base_headers,
                **headers_overrides,
            },
        )

    return _helper


@pytest.fixture
def rs256_jwks_uri(rs256_domain):
    """
    Provide a fixture that returns a jwks uri for use in other fixtures.

    Args:
        rs256_jwks_uri: An implicit fixture parameter.
    """
    return f"https://{rs256_domain}/.well-known/jwks.json"


@pytest.fixture
def rs256_openid_config(rs256_iss, rs256_jwks_uri):
    """
    Provide a fixture that returns an openid configuration for use in other fixtures.

    Args:
        rs256_iss:      An implicit fixture parameter.
        rs256_jwks_uri: An implicit fixture parameter.
    """
    return OpenidConfig(
        issuer=rs256_iss,
        jwks_uri=rs256_jwks_uri,
    )


def build_mock_openid_server(
    domain, openid_config, jwk, jwks_uri
) -> Callable[[str, OpenidConfig, JWK, str], _GeneratorContextManager]:
    """
    Provide a fixture that returns a context manager that mocks opend-config routes.

    Args:
        domain:        The domain of the openid server to mock.
        openid_config: The config to return from the mocked config route.
        jwk:           The jwk to return from the mocked jwk route.
        jwks_uri:      The URL of the jwks route to mock.

    Returns:
        A context manager that, while active, mocks the openid routes needed by Armasec.
    """

    @contextmanager
    def _helper(
        domain: str = domain,
        openid_config: OpenidConfig = openid_config,
        jwk: JWK = jwk,
        jwks_uri: str = jwks_uri,
    ):
        MockOpenidRoutes = namedtuple("MockOpenidRoutes", ["openid_config_route", "jwks_route"])
        with respx.mock:
            openid_config_route = respx.get(
                OpenidConfigLoader.build_openid_config_url(domain),
            )
            openid_config_route.return_value = httpx.Response(
                starlette.status.HTTP_200_OK,
                json=openid_config.model_dump(mode="json"),
            )

            jwks = JWKs(keys=[jwk])
            jwks_route = respx.get(jwks_uri)
            jwks_route.return_value = httpx.Response(
                starlette.status.HTTP_200_OK,
                json=jwks.model_dump(mode="json"),
            )
            yield MockOpenidRoutes(openid_config_route, jwks_route)

    return _helper


@pytest.fixture
def mock_openid_server(rs256_domain, rs256_openid_config, rs256_jwk, rs256_jwks_uri):
    """
    Provide a fixture that mocks an openid server using the extension fixtures.

    Args:
        rs256_domain:        An implicit fixture parameter.
        rs256_openid_config: An implicit fixture parameter.
        rs256_jwk:           An implicit fixture parameter.
        rs256_jwks_uri:      An implicit fixture parameter.
    """
    builder = build_mock_openid_server(rs256_domain, rs256_openid_config, rs256_jwk, rs256_jwks_uri)
    with builder() as constructed_builder:
        yield constructed_builder



"""
This module provides an abstract base class for algorithmic token decoders
"""

from __future__ import annotations

from functools import partial
from typing import Callable

import jmespath
import buzz
from jose import jwt

from armasec.exceptions import AuthenticationError, PayloadMappingError
from armasec.schemas.jwks import JWKs
from armasec.token_payload import TokenPayload
from armasec.utilities import log_error, noop


class TokenDecoder:
    """
    Decoder class used to decode tokens given an algorithm and jwks.
    """

    algorithm: str

    def __init__(
        self,
        jwks: JWKs,
        algorithm: str = "RS256",
        debug_logger: Callable[..., None] | None = None,
        decode_options_override: dict | None = None,
        payload_claim_mapping: dict | None = None,
    ):
        """
        Initializes a TokenDecoder.

        Args:
            algorithm:               The algorithm to use for decoding. Defaults to RS256.
            jwks:                    JSON web keys object holding the public keys for decoding.
            openid_config:           The openid_configuration needed for claims such as 'issuer'.
            debug_logger:            A callable, that if provided, will allow debug logging. Should
                                     be passed as a logger method like `logger.debug`
            decode_options_override: Options that can override the default behavior of the jwt
                                     decode method. For example, one can ignore token expiration by
                                     setting this to `{ "verify_exp": False }`
            payload_claim_mapping:   Optional mappings that are applied to map claims to top-level
                                     attribute of TokenPayload using a dict format of:

                                     ```
                                     {
                                         "top_level_attribute": "decoded.token.JMESPath"
                                     }
                                     ```
                                     The values _must_ be a valid JMESPath.

                                     Consider this example:

                                     ```
                                     {
                                          "permissions": "resource_access.default.roles"
                                     }
                                     ```

                                     The above example would result in a TokenPayload like:

                                     ```
                                     TokenPayload(permissions=token["resource_access"]["default"]["roles"])
                                     ```
                                     Raises a 500 if the path does not match
        """
        self.algorithm = algorithm
        self.jwks = jwks
        self.debug_logger = debug_logger if debug_logger else noop
        self.decode_options_override = decode_options_override if decode_options_override else {}
        self.payload_claim_mapping = payload_claim_mapping if payload_claim_mapping else {}

    def get_decode_key(self, token: str) -> dict:
        """
        Search for a public keys within the JWKs that matches the incoming token.

        Compares the token's unverified header against available JWKs. Uses the matching JWK for the
        decode key.  Raise AuthenticationError if matching public key cannot be found.

        Args:
            token: The token to match against available JWKs.
        """
        self.debug_logger("Getting decode key from JWKs")
        unverified_header = jwt.get_unverified_header(token)
        self.debug_logger(f"Extraced unverified header: {unverified_header}")
        kid = unverified_header.get("kid")
        AuthenticationError.require_condition(
            kid,
            "Unverified header doesn't contain 'kid'...not sure how this happened",
        )

        for jwk in self.jwks.keys:
            self.debug_logger(f"Checking key in jwk: {jwk}")
            if jwk.kid == kid:
                self.debug_logger("Key matches unverified header. Using as decode secret.")
                return jwk.model_dump()

        raise AuthenticationError("Could not find a matching jwk")

    def decode(self, token: str, **claims) -> TokenPayload:
        """
        Decode a JWT into a TokenPayload while checking signatures and claims.

        Args:
            token:  The token to decode.
            claims: Additional claims to verify in the token.
        """
        self.debug_logger(f"Attempting to decode '{token}'")
        self.debug_logger(f"  checking claims: {claims}")
        with AuthenticationError.handle_errors(
            "Failed to decode token string",
            do_except=partial(log_error, self.debug_logger),
        ):
            payload_dict = dict(
                jwt.decode(
                    token,
                    self.get_decode_key(token),
                    algorithms=[self.algorithm],
                    options=self.decode_options_override,
                    **claims,
                )
            )
            self.debug_logger(f"Raw payload dictionary is {payload_dict}")

        with PayloadMappingError.handle_errors(
            "Failed to map decoded token to payload",
            do_except=partial(log_error, self.debug_logger),
        ):
            for payload_key, token_jmespath in self.payload_claim_mapping.items():
                mapped_value = jmespath.search(token_jmespath, payload_dict)
                buzz.require_condition(
                    mapped_value is not None,
                    f"No matching values found for claim mapping {token_jmespath} -> {payload_key}",
                    raise_exc_class=KeyError,
                )
                payload_dict[payload_key] = mapped_value
            self.debug_logger(f"Mapped payload dictionary is {payload_dict}")

            self.debug_logger("Attempting to convert to TokenPayload")
            token_payload = TokenPayload(
                **payload_dict,
                original_token=token,
            )
            self.debug_logger(f"Built token_payload as {token_payload}")
            return token_payload



from sys import exc_info
from functools import wraps

import buzz
import snick
import typer
from loguru import logger
from rich import traceback
from rich.console import Console
from rich.panel import Panel


# Enables prettified traceback printing via rich
traceback.install()


class ArmasecCliError(buzz.Buzz):
    pass


class Abort(buzz.Buzz):
    def __init__(
        self,
        message,
        *args,
        subject=None,
        log_message=None,
        warn_only=False,
        **kwargs,
    ):
        self.subject = subject
        self.log_message = log_message
        self.warn_only = warn_only
        (_, self.original_error, __) = exc_info()
        super().__init__(message, *args, **kwargs)


def handle_abort(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Abort as err:
            if not err.warn_only:
                if err.log_message is not None:
                    logger.error(err.log_message)

                if err.original_error is not None:
                    logger.error(f"Original exception: {err.original_error}")

            panel_kwargs = dict()
            if err.subject is not None:
                panel_kwargs["title"] = f"[red]{err.subject}"
            message = snick.dedent(err.message)

            console = Console()
            console.print()
            console.print(Panel(message, **panel_kwargs))
            console.print()
            raise typer.Exit(code=1)

    return wrapper



from typing import Optional

import httpx
from pydantic import BaseModel

from armasec_cli.config import Settings


class TokenSet(BaseModel):
    access_token: str
    refresh_token: Optional[str] = None


class IdentityData(BaseModel):
    client_id: str
    email: Optional[str] = None


class Persona(BaseModel):
    token_set: TokenSet
    identity_data: IdentityData


class DeviceCodeData(BaseModel):
    device_code: str
    verification_uri_complete: str
    interval: int


class CliContext(BaseModel, arbitrary_types_allowed=True):
    persona: Optional[Persona] = None
    client: Optional[httpx.Client] = None
    settings: Optional[Settings] = None



from __future__ import annotations

from time import sleep
from typing import cast

import snick
from jose.exceptions import ExpiredSignatureError
from loguru import logger
from jose import jwt
from pydantic import ValidationError

from armasec_cli.exceptions import Abort, ArmasecCliError
from armasec_cli.schemas import DeviceCodeData, TokenSet, IdentityData
from armasec_cli.cache import load_tokens_from_cache, save_tokens_to_cache
from armasec_cli.client import make_request
from armasec_cli.config import OidcProvider
from armasec_cli.schemas import Persona, CliContext
from armasec_cli.format import terminal_message
from armasec_cli.time_loop import TimeLoop


def extract_persona(token_set: TokenSet | None = None):
    if token_set is None:
        token_set = load_tokens_from_cache()

    try:
        identity_data = validate_token_and_extract_identity(token_set)
    except ExpiredSignatureError:
        Abort.require_condition(
            token_set.refresh_token is not None,
            "The auth token is expired. Please retrieve a new and log in again.",
            raise_kwargs=dict(
                subject="Expired access token",
            ),
        )

        logger.debug("The access token is expired. Attempting to refresh token")
        refresh_access_token(token_set)
        identity_data = validate_token_and_extract_identity(token_set)

    logger.debug(f"Persona created with identity_data: {identity_data}")

    save_tokens_to_cache(token_set)

    return Persona(
        token_set=token_set,
        identity_data=identity_data,
    )


def validate_token_and_extract_identity(token_set: TokenSet) -> IdentityData:
    logger.debug("Validating access token")

    token_file_is_empty = not token_set.access_token
    if token_file_is_empty:
        logger.debug("Access token file exists but it is empty")
        raise Abort(
            """
            Access token file exists but it is empty.

            Please try logging in again.
            """,
            subject="Empty access token file",
            log_message="Empty access token file",
        )

    with Abort.handle_errors(
        """
        There was an unknown error while validating the access token.

        Please try logging in again.
        """,
        ignore_exc_class=ExpiredSignatureError,  # Will be handled in calling context
        raise_kwargs=dict(
            subject="Invalid access token",
            log_message="Unknown error while validating access access token",
        ),
    ):
        token_data = jwt.decode(
            token_set.access_token,
            None,
            options=dict(
                verify_signature=False,
                verify_aud=False,
                verify_exp=True,
            ),
        )

    logger.debug("Extracting identity data from the access token")
    with Abort.handle_errors(
        """
        There was an error extracting the user's identity from the access token.

        Please try logging in again.
        """,
        handle_exc_class=ValidationError,
        raise_kwargs=dict(
            subject="Missing user data",
            log_message="Token data could not be extracted to identity",
        ),
    ):
        identity = IdentityData(
            email=token_data.get("email"),
            client_id=token_data.get("azp"),
        )

    return identity


def init_persona(ctx: CliContext, token_set: TokenSet | None = None):
    if token_set is None:
        token_set = load_tokens_from_cache()

    try:
        identity_data = validate_token_and_extract_identity(token_set)
    except ExpiredSignatureError:
        Abort.require_condition(
            token_set.refresh_token is not None,
            "The auth token is expired. Please retrieve a new and log in again.",
            raise_kwargs=dict(
                subject="Expired access token",
            ),
        )

        logger.debug("The access token is expired. Attempting to refresh token")
        refresh_access_token(ctx, token_set)
        identity_data = validate_token_and_extract_identity(token_set)

    logger.debug(f"Persona created with identity_data: {identity_data}")

    save_tokens_to_cache(token_set)

    return Persona(
        token_set=token_set,
        identity_data=identity_data,
    )


def refresh_access_token(ctx: CliContext, token_set: TokenSet):
    """
    Attempt to fetch a new access token given a refresh token in a token_set.

    Sets the access token in-place.

    If refresh fails, notify the user that they need to log in again.
    """
    print("MAKE THIS FUCKING THING USE THE BASE URL")
    url = "/protocol/openid-connect/token"
    logger.debug(f"Requesting refreshed access token from {url}")

    refreshed_token_set: TokenSet = cast(
        TokenSet,
        make_request(
            ctx.client,
            "/protocol/openid-connect/token",
            "POST",
            abort_message="The auth token could not be refreshed. Please try logging in again.",
            abort_subject="EXPIRED ACCESS TOKEN",
            response_model_cls=TokenSet,
            data=dict(
                client_id=ctx.settings.oidc_client_id,
                audience=ctx.settings.oidc_audience,
                grant_type="refresh_token",
                refresh_token=token_set.refresh_token,
            ),
        ),
    )

    token_set.access_token = refreshed_token_set.access_token


def fetch_auth_tokens(ctx: CliContext) -> TokenSet:
    """
    Fetch an access token (and possibly a refresh token) from Auth0.

    Prints out a URL for the user to use to authenticate and polls the token endpoint to fetch it
    when the browser-based process finishes.
    """
    if ctx.settings.oidc_provider == OidcProvider.KEYCLOAK:
        device_path = "/protocol/openid-connect/auth/device"
        token_path = "/protocol/openid-connect/token"
    elif ctx.settings.oidc_provider == OidcProvider.AUTH0:
        device_path = "/oauth/device/code"
        token_path = "oauth/token"
    else:
        raise ArmasecCliError("Unsupported OIDC Provider.")

    device_code_data: DeviceCodeData = cast(
        DeviceCodeData,
        make_request(
            ctx.client,
            device_path,
            "POST",
            expected_status=200,
            abort_message=(
                """
                There was a problem retrieving a device verification code from
                the auth provider
                """
            ),
            abort_subject="COULD NOT RETRIEVE TOKEN",
            response_model_cls=DeviceCodeData,
            data=dict(
                client_id=ctx.settings.oidc_client_id,
                grant_type="client_credentials",
                audience=ctx.settings.oidc_audience,
            ),
        ),
    )

    max_poll_time = 5 * 60  # 5 minutes
    terminal_message(
        f"""
        To complete login, please open the following link in a browser:

          {device_code_data.verification_uri_complete}

        Waiting up to {max_poll_time / 60} minutes for you to complete the process...
        """,
        subject="Waiting for login",
    )

    for tick in TimeLoop(
        ctx.settings.oidc_max_poll_time,
        message="Waiting for web login",
    ):
        response_data: dict = cast(
            dict,
            make_request(
                ctx.client,
                token_path,
                "POST",
                abort_message=snick.unwrap(
                    """
                    There was a problem retrieving a device verification code
                    from the auth provider
                    """
                ),
                abort_subject="COULD NOT FETCH ACCESS TOKEN",
                data=dict(
                    grant_type="urn:ietf:params:oauth:grant-type:device_code",
                    device_code=device_code_data.device_code,
                    client_id=ctx.settings.oidc_client_id,
                ),
            ),
        )
        if "error" in response_data:
            if response_data["error"] == "authorization_pending":
                logger.debug(f"Token fetch attempt #{tick.counter} failed")
                logger.debug(f"Will try again in {device_code_data.interval} seconds")
                sleep(device_code_data.interval)
            else:
                # TODO: Test this failure condition
                raise Abort(
                    snick.unwrap(
                        """
                        There was a problem retrieving a device verification code
                        from the auth provider:
                        Unexpected failure retrieving access token.
                        """
                    ),
                    subject="Unexpected error",
                    log_message=f"Unexpected error response: {response_data}",
                )
        else:
            return TokenSet(**response_data)

    raise Abort(
        "Login process was not completed in time. Please try again.",
        subject="Timed out",
        log_message="Timed out while waiting for user to complete login",
    )



"""
This module provides the OpenidConfigLoader which is used to load openid-configuration data from an
OIDC provider.
"""

from functools import partial
from typing import Callable, Optional

import httpx
import starlette

from armasec.exceptions import AuthenticationError
from armasec.schemas.jwks import JWKs
from armasec.schemas.openid_config import OpenidConfig
from armasec.utilities import log_error, noop


class OpenidConfigLoader:
    _config: Optional[OpenidConfig] = None
    _jwks: Optional[JWKs] = None

    def __init__(
        self,
        domain: str,
        use_https: bool = True,
        debug_logger: Optional[Callable[..., None]] = None,
    ):
        """
        Initializes a base TokenManager.

        Args:
            secret:        The secret key needed to decode a token
            domain:        The domain of the OIDC provider. This is to construct the
                           openid-configuration url
            use_https:     If falsey, use ``http`` instead of ``https`` (the default).
            debug_logger:  A callable, that if provided, will allow debug logging. Should be passed
                           as a logger method like `logger.debug`
        """
        self.domain = domain
        self.use_https = use_https
        self.debug_logger = debug_logger if debug_logger else noop

    @staticmethod
    def build_openid_config_url(domain: str, use_https: bool = True):
        """
        Builds a url for an openid configuration given a domain.

        Args:
            domain:    The domain of the OIDC provider for which to build a URL
            use_https: Use `https` for the URL by default. If falsey, use `http` instead.
        """
        protocol = "https" if use_https else "http"
        return f"{protocol}://{domain}/.well-known/openid-configuration"

    def _load_openid_resource(self, url: str):
        """
        Helper method to load data from an openid connect resource.
        """
        self.debug_logger(f"Attempting to fetch from openid resource '{url}'")
        with AuthenticationError.handle_errors(
            message=f"Call to url {url} failed",
            do_except=partial(log_error, self.debug_logger),
        ):
            response = httpx.get(url)
        AuthenticationError.require_condition(
            response.status_code == starlette.status.HTTP_200_OK,
            f"Didn't get a success status code from url {url}: {response.status_code}",
        )
        return response.json()

    @property
    def config(self) -> OpenidConfig:
        """
        Retrive the openid config from an OIDC provider. Lazy loads the config so that API calls are
        deferred until the coniguration is needed.
        """
        if not self._config:
            self.debug_logger("Fetching openid configration")
            data = self._load_openid_resource(
                self.build_openid_config_url(self.domain, self.use_https)
            )
            with AuthenticationError.handle_errors(
                message="openid config data was invalid",
                do_except=partial(log_error, self.debug_logger),
            ):
                self._config = OpenidConfig(**data)

        return self._config

    @property
    def jwks(self) -> JWKs:
        """
        Retrives JWKs public keys from an OIDC provider. Lazy loads the jwks so that API calls are
        deferred until the jwks are needed.
        """
        if not self._jwks:
            self.debug_logger("Fetching jwks")
            data = self._load_openid_resource(str(self.config.jwks_uri))
            with AuthenticationError.handle_errors(
                message="jwks data was invalid",
                do_except=partial(log_error, self.debug_logger),
            ):
                self._jwks = JWKs(**data)

        return self._jwks



from __future__ import annotations

from functools import wraps
from pathlib import Path
from typing import Any, TypeVar

import httpx
import pydantic
import snick
import typer
from loguru import logger

from armasec_cli.exceptions import Abort
from armasec_cli.config import Settings


def build_client(settings: Settings):
    protocol = "https" if settings.oidc_use_https else "http"
    base_url = f"{protocol}://{settings.oidc_domain}"
    logger.debug(f"Creating client with base URL {base_url}")
    return httpx.Client(
        base_url=base_url,
        headers={"content-type": "application/x-www-form-urlencoded"},
    )


def attach_client(func):
    @wraps(func)
    def wrapper(ctx: typer.Context, *args, **kwargs):
        if ctx.obj.settings is None:
            raise Abort(
                """
                Cannot attach client before settings!

                Check the order of the decorators and make sure settings are attached
                first.
                """,
                subject="Attaching out of order",
                log_message="Attaching out of order",
            )

        logger.debug("Binding client to CLI context")
        ctx.obj.client = build_client(ctx.obj.settings)
        return func(ctx, *args, **kwargs)

    return wrapper


def _deserialize_request_model(
    request_model: pydantic.BaseModel,
    request_kwargs: dict[str, Any],
    abort_message: str,
    abort_subject: str,
):
    """
    Deserialize a pydantic model instance into request_kwargs for a request in place.
    """
    Abort.require_condition(
        all(
            [
                "data" not in request_kwargs,
                "json" not in request_kwargs,
                "content" not in request_kwargs,
            ]
        ),
        snick.unwrap(
            f"""
            {abort_message}:
            Request was incorrectly structured.
            """
        ),
        raise_kwargs=dict(
            subject=abort_subject,
            log_message=snick.unwrap(
                """
                When using `request_model`, you may not pass
                `data`, `json`, or `content` in the `request_kwargs`
                """
            ),
        ),
    )
    with Abort.handle_errors(
        snick.unwrap(
            f"""
            {abort_message}:
            Request data could not be deserialized for http request.
            """
        ),
        raise_kwargs=dict(
            subject=abort_subject,
            log_message=snick.unwrap(
                f"""
                Could not deserialize instance of {request_model.__class__}:
                {request_model}
                """
            ),
        ),
    ):
        request_kwargs["content"] = request_model.model_dump_json()
        request_kwargs["headers"] = {"Content-Type": "application/json"}


ResponseModel = TypeVar("ResponseModel", bound=pydantic.BaseModel)


def make_request(
    client: httpx.Client,
    url_path: str,
    method: str,
    *,
    expected_status: int | None = None,
    expect_response: bool = True,
    abort_message: str = "There was an error communicating with the API",
    abort_subject: str = "REQUEST FAILED",
    response_model_cls: type[ResponseModel] | None = None,
    request_model: pydantic.BaseModel | None = None,
    save_to_file: Path | None = None,
    **request_kwargs,
) -> ResponseModel | dict | int:
    """
    Make a request against the Jobbergate API.

    Args:
        url_path:            The path to add to the base url of the client where the
                             request should be sent
        method:              The REST method to use for the request
                             (GET, PUT, UPDATE, POST, DELETE, etc)
        expected_status:     The status code to expect on the response. If it is not
                             received, raise an Abort
        expect_response:     Indicates if response data (JSON) is expected from the API
                             endpoint
        abort_message:       The message to show the user if there is a problem and the
                             app must be aborted
        abort_subject:       The subject to use in Abort output to the user
        response_model_cls:  If supplied, serialize the response data into this Pydantic
                             model class
        request_model:       Use a pydantic model instance as the data body for the
                             request
        request_kwargs:      Any additional keyword arguments that need to be passed on
                             to the client
    """

    if request_model is not None:
        _deserialize_request_model(
            request_model,
            request_kwargs,
            abort_message,
            abort_subject,
        )

    logger.debug(f"Making request to url_path={url_path}")
    request = client.build_request(method, url_path, **request_kwargs)

    # Look for the request body in the request_kwargs
    debug_request_body = request_kwargs.get(
        "data",
        request_kwargs.get("json", request_kwargs.get("content")),
    )
    logger.debug(
        snick.dedent(
            f"""
            Request built with:
              url:     {request.url}
              method:  {method}
              headers: {request.headers}
              body:    {debug_request_body}
            """
        )
    )

    with Abort.handle_errors(
        snick.unwrap(
            f"""
            {abort_message}:
            Communication with the API failed.
            """
        ),
        raise_kwargs=dict(
            subject=abort_subject,
            log_message="There was an error making the request to the API",
        ),
    ):
        response = client.send(request)

    if expected_status is not None and response.status_code != expected_status:
        if (
            method in ("PATCH", "PUT", "DELETE")
            and response.status_code == 403
            and "does not own" in response.text
        ):
            raise Abort(
                snick.dedent(
                    f"""
                     {abort_message}:
                     [red]You do not own this resource.[/red]
                     Please contact the owner if you need it to be modified.
                     """,
                ),
                subject=abort_subject,
                log_message=snick.unwrap(
                    f"""
                    Resource could not be modified by non-owner:
                    {response.text}
                    """
                ),
            )
        else:
            raise Abort(
                snick.unwrap(
                    f"""
                    {abort_message}:
                    Received an error response.
                    """
                ),
                subject=abort_subject,
                log_message=snick.unwrap(
                    f"""
                    Got an error code for request:
                    {response.status_code}:
                    {response.text}
                    """
                ),
            )

    if save_to_file is not None:
        save_to_file.parent.mkdir(parents=True, exist_ok=True)
        save_to_file.write_bytes(response.content)
        return response.status_code

    # TODO: constrain methods with a named enum
    if expect_response is False or method == "DELETE":
        return response.status_code

    with Abort.handle_errors(
        snick.unwrap(
            f"""
            {abort_message}:
            Response carried no data.
            """
        ),
        raise_kwargs=dict(
            subject=abort_subject,
            log_message=f"Failed unpacking json: {response.text}",
        ),
    ):
        data = response.json()
    logger.debug(f"Extracted data from response: {data}")

    if response_model_cls is None:
        return data

    logger.debug("Validating response data with ResponseModel")
    with Abort.handle_errors(
        snick.unwrap(
            f"""
            {abort_message}:
            Unexpected data in response.
            """
        ),
        raise_kwargs=dict(
            subject=abort_subject,
            log_message=f"Unexpected format in response data: {data}",
        ),
    ):
        return response_model_cls(**data)



"""
Manage plugins from armasec.
"""

import pluggy
from armasec.pluggable import hookspecs

hookimpl = pluggy.HookimplMarker("armasec")
plugin_manager = pluggy.PluginManager("armasec")
plugin_manager.add_hookspecs(hookspecs)
plugin_manager.load_setuptools_entrypoints("armasec")
